/**
 * bicg.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

#include "scalapack_util.h"
#define BLOCK_SIZE 100

#define NX 30000
#define NY 30000

/* Include benchmark-specific header. */
/* Default data type is double, default size is 4000. */
#include "bicg.h"


/* Array initialization. */
static
void init_array (int nx, int ny,
		 DATA_TYPE POLYBENCH_2D(A,NX,NY,nx,ny),
		 DATA_TYPE POLYBENCH_1D(r,NX,nx),
		 DATA_TYPE POLYBENCH_1D(p,NY,ny))
{
  int i, j;

  for (i = 0; i < ny; i++)
    p[i] = i * M_PI;
  for (i = 0; i < nx; i++) {
    r[i] = i * M_PI;
    for (j = 0; j < ny; j++)
      A[i][j] = ((DATA_TYPE) j*(i+1))/nx;
  }
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int nx, int ny,
		 DATA_TYPE POLYBENCH_1D(s,NY,ny),
		 DATA_TYPE POLYBENCH_1D(q,NX,nx))

{
  int i;

  for (i = 0; i < ny; i++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, s[i]);
    if (i % 20 == 0) fprintf (stderr, "\n");
  }
  for (i = 0; i < nx; i++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, q[i]);
    if (i % 20 == 0) fprintf (stderr, "\n");
  }
  fprintf (stderr, "\n");
}

#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,NX,NY,nx,ny)
#define __DECLARATION_OF_s DATA_TYPE POLYBENCH_1D(s,NY,ny)
#define __DECLARATION_OF_q DATA_TYPE POLYBENCH_1D(q,NX,nx)
#define __DECLARATION_OF_p DATA_TYPE POLYBENCH_1D(p,NY,ny)
#define __DECLARATION_OF_r DATA_TYPE POLYBENCH_1D(r,NX,nx)

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_bicg(int nx, int ny,
		 DATA_TYPE POLYBENCH_2D(A,NX,NY,nx,ny),
		 DATA_TYPE POLYBENCH_1D(s,NY,ny),
		 DATA_TYPE POLYBENCH_1D(q,NX,nx),
		 DATA_TYPE POLYBENCH_1D(p,NY,ny),
		 DATA_TYPE POLYBENCH_1D(r,NX,nx))
{
  int i, j;

#pragma scop
  for (i = 0; i < _PB_NY; i++)
    s[i] = 0;
  for (i = 0; i < _PB_NX; i++)
    {
      q[i] = 0;
      for (j = 0; j < _PB_NY; j++)
	{
	  s[j] = s[j] + r[i] * A[i][j];
	  q[i] = q[i] + A[i][j] * p[j];
	}
    }
#pragma endscop

}

/* Main scalapack computational kernel. The whole function will be timed,
   including the call and return. */
static
void scalapack_kernel_bicg(int nx, int ny,
         int nprocs)
{
  int i, j;

double *A_local, *p_local, *q_local, *r_local, *s_local;

MDESC desc_A, desc_p, desc_q, desc_r, desc_s;
MDESC desc_A_local, desc_p_local, desc_q_local, desc_r_local, desc_s_local;

MKL_INT iam,  ictxt, myrow, mycol, nprow, npcol;
MKL_INT lld, lld_local, info, nx_mkl, ny_mkl, nb, nq, mp, nq_vec, mp_vec;

    int n_procs_row = 1, n_procs_col = 1;

    get_procs_grid_size(nprocs , &n_procs_row, &n_procs_col);

    nprow = (MKL_INT) n_procs_row;
    npcol = (MKL_INT) n_procs_col;

    nx_mkl = (MKL_INT) nx;
    ny_mkl = (MKL_INT) ny;

    nb = (MKL_INT) BLOCK_SIZE;

/*  Init working 2D process grid */
    blacs_get_( &i_negone, &i_zero, &ictxt );
    blacs_gridinit_( &ictxt, "R", &nprow, &npcol );
    blacs_gridinfo_( &ictxt, &nprow, &npcol, &myrow, &mycol );

/*  Compute precise length of local pieces and allocate array on
    each process for parts of distributed vectors */
    mp = numroc_( &nx_mkl, &nb, &myrow, &i_zero, &nprow );
    nq = numroc_( &ny_mkl, &nb, &mycol, &i_zero, &npcol );

    mp_vec = numroc_( &i_one, &i_one, &myrow, &i_zero, &nprow );
    nq_vec = numroc_( &i_one, &i_one, &mycol, &i_zero, &npcol );
    printf("myrow=%d, mycol=%d mp=%d, nq = %d, nq_rhs = %d\n", myrow, mycol, mp, nq, nq_vec);
    if( ( myrow == 0 ) && ( mycol == 0 ) ){
        my_rank = 0;
    }

    A_local = (double*) calloc( mp*nq, sizeof( double ) );

/*  Only #0 process column contains distributed 1-d vectors*/
    if ( nq_vec != 0 ){
        q_local = (double *) calloc(mp, sizeof( double ));
        r_local = (double *) calloc(mp, sizeof( double ));
        p_local = (double *) calloc(nq, sizeof( double ));
        s_local = (double *) calloc(nq, sizeof( double ));
    } else {                                                                                 
        p_local = NULL;
        q_local = NULL;
        r_local = NULL;
        s_local = NULL;
    }

   /*  Compute leading dimensions */
    lld_local = MAX( numroc_( &nx_mkl, &ny_mkl, &myrow, &i_zero, &nprow ), 1 );
    lld = MAX( mp, 1 );

    /*  Initialize descriptors for initial arrays located on 0 process */
    descinit_( desc_A, &nx_mkl, &ny_mkl, &nx_mkl, &ny_mkl, &i_zero, &i_zero, &ictxt, &lld_local, &info );

    descinit_( desc_q, &nx_mkl, &i_one, &nx_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_r, &nx_mkl, &i_one, &nx_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_s, &ny_mkl, &i_one, &ny_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_p, &ny_mkl, &i_one, &ny_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );


/*  Initialize descriptors for distributed arrays */

    descinit_( desc_A_local, &nx_mkl, &ny_mkl, &nb, &nb, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_q_local, &nx_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_r_local, &nx_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_p_local, &ny_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_s_local, &nx_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );


/*  Distribute vectors and matrices from 0 process over process grid */
#ifdef USE_LOCAL_ARRAYS
        generate_random_values(A_local, mp*nq);
        generate_random_values(q_local, mp);
        generate_random_values(r_local, mp);
        generate_random_values(s_local, nq);
        generate_random_values(p_local, nq);
 
#else
      /* Variable declaration/allocation. */
      POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, NX, NY, nx, ny);
      POLYBENCH_1D_ARRAY_DECL(s, DATA_TYPE, NY, ny);
      POLYBENCH_1D_ARRAY_DECL(p, DATA_TYPE, NY, ny);
      POLYBENCH_1D_ARRAY_DECL(q, DATA_TYPE, NX, nx);
      POLYBENCH_1D_ARRAY_DECL(r, DATA_TYPE, NX, nx);

      /* Initialize array(s). */
      init_array (nx, ny,
	      POLYBENCH_ARRAY(A),
	      POLYBENCH_ARRAY(r),
	      POLYBENCH_ARRAY(p));


       pdgeadd_( &trans_N, &nx_mkl, &ny_mkl, &one, 
            A, &i_one, &i_one, desc_A,
            &zero, 
            A_local, &i_one, &i_one, desc_A_local );


    pdgeadd_( &trans_N, &nx_mkl, &i_one, &one, r, &i_one, 
        &i_one, desc_r, &zero, r_local, &i_one, &i_one, desc_r_local );
    pdgeadd_( &trans_N, &ny_mkl, &i_one, &one, p, &i_one,
        &i_one, desc_p, &zero, p_local, &i_one, &i_one, desc_p_local );
    
#endif

  /* Start timer. */
  polybench_start_instruments;

      // q = A * p
    pdgemv_(&trans_N, &nx_mkl, &ny_mkl, &one, 
            A_local, &i_one, &i_one, desc_A_local,
            p_local, &i_one, &i_one, desc_p_local, &i_one,
            &one,
            q_local, &i_one, &i_one, desc_q_local, &i_one);

      // s = A' * r
    pdgemv_(&trans_T, &nx_mkl, &ny_mkl, &one, 
            A_local, &i_one, &i_one, desc_A_local,
            r_local, &i_one, &i_one, desc_r_local, &i_one, 
            &one,
            s_local, &i_one, &i_one, desc_s_local, &i_one);

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;


#ifndef USE_LOCAL_ARRAYS
    pdgeadd_(&trans_N, &ny_mkl ,&i_one, &one, 
            s_local, &i_one, &i_one, desc_s_local,
            &zero,
            s, &i_one, &i_one, desc_s);

    pdgeadd_(&trans_N, &nx_mkl ,&i_one, &one, 
            q_local, &i_one, &i_one, desc_q_local,
            &zero,
            q, &i_one, &i_one, desc_q);

if(my_rank == 0) {
      /* Prevent dead-code elimination. All live-out data must be printed
         by the function call in argument. */
      polybench_prevent_dce(print_array(nx, ny, POLYBENCH_ARRAY(s), POLYBENCH_ARRAY(q)));

}
      /* Be clean. */
      POLYBENCH_FREE_ARRAY(A);
      POLYBENCH_FREE_ARRAY(s);
      POLYBENCH_FREE_ARRAY(q);
      POLYBENCH_FREE_ARRAY(p);
      POLYBENCH_FREE_ARRAY(r);


#endif

    blacs_gridexit_( &ictxt );
    blacs_exit_( &i_zero );

}
int main(int argc, char** argv)
{
  
  int nx, ny;
  int num_procs = atoi(argv[1]);

#ifndef USE_LOCAL_ARRAYS
      nx = NX;
      ny = NY;
#else
      nx = sqrt(num_procs) * NX;
      ny = sqrt(num_procs) * NY;
#endif

  /* Run kernel. */
  scalapack_kernel_bicg (nx, ny,
           num_procs);

  return 0;
}
