#
# common Makefile
#
# Gets included after the local Makefile in an example sub-directory
#
BASEDIR=$(dir $(lastword $(MAKEFILE_LIST)))

CC=icc
#CC=gcc

#MPI=openmpi
MPI=mvapich
#MPI=intelmpi

NPROCS=2
NTHREADS=1
NTHREADS_WITH_MPI=2
NTHREADS_LIST = 1 2 4 8 16 32
NPROCS_LIST = 1 2 4 8 16 32
TILE_SIZES = 16 32 64 128 256
POLYRTINCDIR=../polyrt
POLYRTLIBDIR=../polyrt/.libs
POLYBENCHINCDIR=$(BASEDIR)polybench/utilities
POLYBENCHSRC=$(BASEDIR)polybench/utilities/polybench.c
HOSTS_FILE=$(BASEDIR)hosts

PLC=$(BASEDIR)../polycc

# Intel MKL and AMD ACML library paths
MKL=/opt/intel/mkl
ACML=/usr/local/acml

#Metis libray path
METIS=/usr/local/lib/libmetis.a

#Intel scalapack library paths and compiler flags
S_CFLAGS +=  -DMKL_ILP64 -I$(MKLROOT)/include 
S_LFLAGS +=  $(MKLROOT)/lib/intel64/libmkl_scalapack_ilp64.a -Wl,--start-group  $(MKLROOT)/lib/intel64/libmkl_cdft_core.a $(MKLROOT)/lib/intel64/libmkl_intel_ilp64.a $(MKLROOT)/lib/intel64/libmkl_intel_thread.a $(MKLROOT)/lib/intel64/libmkl_core.a $(MKLROOT)/lib/intel64/libmkl_blacs_intelmpi_ilp64.a -Wl,--end-group -lpthread -lm  


ifeq ($(CC), icc)
	CXX           := icpc
	OPT_FLAGS     := -O3 -fp-model precise -ansi-alias -ipo
	PAR_FLAGS     := -parallel
	OMP_FLAGS     := -openmp
	ifeq ($(MPI), mvapich)
		MPICC        := mpicc -cc=icc -D__MPI
		MPICXX       := mpicc -cc=icpc -D__MPI
	else ifeq ($(MPI), intelmpi)
		MPICC        := mpiicc -D__MPI
		MPICXX       := mpiicpc -D__MPI -DMPICH_IGNORE_CXX_SEEK -DMPICH_SKIP_MPICXX
	else #openmpi
		MPICC        := OMPI_CC=icc mpicc -D__MPI
		MPICXX       := OMPI_CC=icpc mpicc -D__MPI
	endif
else
	# for gcc
	CXX           := g++
	OPT_FLAGS     := -O3 -ftree-vectorize -msse3 
	PAR_FLAGS     := -ftree-parallelize-loops=4
	OMP_FLAGS     := -fopenmp
	ifeq ($(MPI), mvapich)
		MPICC        := mpicc -cc=gcc -D__MPI
		MPICXX       := mpicc -cc=g++ -D__MPI
	else ifeq ($(MPI), intelmpi)
		MPICC        := mpicc -D__MPI
		MPICXX       := mpicxx -D__MPI
	else #openmpi
		MPICC        := OMPI_CC=gcc mpicc -D__MPI
		MPICXX       := OMPI_CC=g++ mpicc -D__MPI
	endif
endif

ifeq ($(NON_CLUSTER), true)
	MPI = openmpi
endif

ifeq ($(MPI), mvapich)
	MPIARGS       = -np $(NPROCS) -hostfile $(HOSTS_FILE)
	MPIENV        = MV2_ENABLE_AFFINITY=0 OMP_NUM_THREADS=$(NTHREADS_WITH_MPI)
	MPIRUN        = mpirun_rsh $(MPIARGS) $(MPIENV)
else ifeq ($(MPI), intelmpi)
	MPIARGS       = -ppn=1 -np $(NPROCS) -hostfile $(HOSTS_FILE)
	MPIENV        = -env I_MPI_FABRICS ofa -env OMP_NUM_THREADS $(NTHREADS_WITH_MPI)
	MPIRUN        = mpirun $(MPIARGS) $(MPIENV)
else #openmpi
	MPIARGS       = -np $(NPROCS)
	MPIENV        = OMP_NUM_THREADS=$(NTHREADS_WITH_MPI)
	MPIRUN        = $(MPIENV) mpirun $(MPIARGS)
endif

CFLAGS += -DTIME
LDFLAGS += -lm
PLCFLAGS += --isldep --lastwriter
DISTOPT_FLAGS += --cloogsh --commreport
TILEFLAGS += 

#PERFCTR=perfctr

ifdef PERFCTR
	CFLAGS += -DPERFCTR -L/usr/local/lib64 -lpapi
endif

ifdef POLYBENCH
	CFLAGS += -DPOLYBENCH_USE_SCALAR_LB -DPOLYBENCH_TIME -I $(POLYBENCHINCDIR) $(POLYBENCHSRC)
	DISTOPT_FLAGS += --variables_not_global
endif

dist_dynsched: $(SRC).dist_dynsched.c sigma_$(SRC).dist_dynsched.c pi_$(SRC).dist_dynsched.c
	$(MPICXX) $(OPT_FLAGS) $(OMP_FLAGS) $(CFLAGS) $(SRC).dist_dynsched.c sigma_$(SRC).dist_dynsched.c pi_$(SRC).dist_dynsched.c \
		$(POLYRTINCDIR)/polyrt.c -o $@ -I $(POLYRTINCDIR) -ltbb $(LDFLAGS) 

dist_dynsched_data_dist: $(SRC).dist_dynsched_data_dist.c sigma_$(SRC).dist_dynsched_data_dist.c pi_$(SRC).dist_dynsched_data_dist.c
	$(MPICXX) $(OPT_FLAGS) $(OMP_FLAGS) -D__DATA_DIST_DECLS $(CFLAGS) $(SRC).dist_dynsched_data_dist.c sigma_$(SRC).dist_dynsched_data_dist.c pi_$(SRC).dist_dynsched_data_dist.c \
		$(POLYRTINCDIR)/polyrt.c $(POLYRTINCDIR)/buffer_manager.c -o $@ -I $(POLYRTINCDIR) -ltbb $(LDFLAGS)

dist_dynsched_data_dist_local: $(SRC).dist_dynsched_data_dist.c sigma_$(SRC).dist_dynsched_data_dist.c pi_$(SRC).dist_dynsched_data_dist.c
	$(MPICXX) $(OPT_FLAGS) $(OMP_FLAGS) $(CFLAGS) -DUSE_LOCAL_ARRAYS $(SRC).dist_dynsched_data_dist.c sigma_$(SRC).dist_dynsched_data_dist.c pi_$(SRC).dist_dynsched_data_dist.c \
		$(POLYRTINCDIR)/polyrt.c $(POLYRTINCDIR)/buffer_manager.c -o $@ -I $(POLYRTINCDIR) -ltbb $(LDFLAGS) 

scalapack: $(SRC).scalapack.c 
	$(MPICC) -DMPI  $(OPT_FLAGS) $(OMP_FLAGS) $(CFLAGS) $(S_CFLAGS) $(SRC).scalapack.c -o scalapack  $(S_LFLAGS) 

scalapack_run: scalapack 
	mpirun_rsh  -np $(NPROCS) -hostfile $(HOSTS_FILE) MV2_ENABLE_AFFINITY=0 OMP_NUM_THREADS=$(NTHREADS) ./scalapack  2>out_scalapack
	touch .test
	rm -f .test

dist_dynsched_data_dist_run: dist_dynsched_data_dist_local
	mpirun_rsh  -np $(NPROCS) -hostfile $(HOSTS_FILE) MV2_ENABLE_AFFINITY=0 OMP_NUM_THREADS=$(NTHREADS_WITH_MPI) ./dist_dynsched_data_dist_local 2> out_dist_dynsched_data_dist_run
	@echo Success!
   
data_dist_compare_scalapack: distopt_data_dist 
	mpirun_rsh  -np $(NPROCS) -hostfile $(HOSTS_FILE) MV2_ENABLE_AFFINITY=0 OMP_NUM_THREADS=$(NTHREADS) ./distopt_data_dist 2> out_distopt_data_dist
	echo "Running distopt data dist for $(SRC) with $(p) procs" >&1
	echo "" >&1 
	make scalapack_run

clean:
	rm -f out_* scalapack dist_dynsched_data_dist_local dist_dynsched_data_dist 

