/**
 * floyd-warshall.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <math.h>
#include <limits.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define min(x,y)    ((x) < (y)? (x) : (y))
#include <upc_relaxed.h>
#include <bupc_collectivev.h>
#include <upc.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

/* Include benchmark-specific header. */
/* Default data type is double, default size is 1024. */
#include "floyd-warshall.h"


/* Array initialization. */
static
void init_array (int n,
                 DATA_TYPE POLYBENCH_2D(path,N/THREADS,N,n/THREADS,n))
{
    int i, j, ii;

    for (i = 0; i < n/THREADS; i++)
        for (j = 0; j < n; j++) {
            ii = (MYTHREAD*n/THREADS)+i;
            path[i][j] = ((DATA_TYPE) (ii+1)*(j+1)) / n;
        }
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
                 DATA_TYPE POLYBENCH_2D(path,N/THREADS,N,n/THREADS,n))

{
    int i, j;

    char name[32];
    strcpy(name, "");
    sprintf(name+strlen(name), "out_%d", MYTHREAD);
    FILE *file = fopen(name, "w");
    for (i = 0; i < n/THREADS; i++) {
        for (j = 0; j < n; j++) {
            fprintf (file, DATA_PRINTF_MODIFIER, path[i][j]);
            //if ((i * n + j) % 20 == 0) fprintf (file, "\n");
            if ((j+1) % 32 == 0) fprintf (file, "\n");
        }
    }
    fprintf (file, "\n");
    fclose(file);
    upc_barrier;
}

#define __DECLARATION_OF_path DATA_TYPE POLYBENCH_2D(path,N,N,n,n)

#define BLOCKSIZE 1024

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_floyd_warshall(int n,
                           DATA_TYPE POLYBENCH_2D(path,N/THREADS,N,n/THREADS,n), shared [BLOCKSIZE] DATA_TYPE globalPath[N], DATA_TYPE localPath[N])
{
    int i, j, k, ownerOfRow, activeRow;

    double t_local = 0.0, t_global = 0.0;

    upc_barrier;
    t_local = rtclock();

    if (0 == MYTHREAD) {
        memcpy(localPath, path[0], _PB_N*sizeof(DATA_TYPE));
        for(i = 0; i < _PB_N; i+=BLOCKSIZE)
            upc_memput(&globalPath[i], &localPath[i], BLOCKSIZE*sizeof(DATA_TYPE));
    }

    upc_barrier;
    if (0 != MYTHREAD) {
        for(i = 0; i < _PB_N; i+=BLOCKSIZE)
            upc_memget(&localPath[i], &globalPath[i], BLOCKSIZE*sizeof(DATA_TYPE));
    }
    upc_barrier;
    for (k = 0; k < _PB_N; k++)
    {
        for(i = 0; i < _PB_N/THREADS; i++)
            for (j = 0; j < _PB_N; j++)
                path[i][j] = path[i][j] < path[i][k] + localPath[j] ?
                             path[i][j] : path[i][k] + localPath[j];

        ownerOfRow = floord((k+1), (_PB_N/THREADS));
        activeRow = k+1 - (ownerOfRow*_PB_N/THREADS);
        if (ownerOfRow == MYTHREAD) {
            memcpy(localPath, path[activeRow], _PB_N*sizeof(DATA_TYPE));
            for(i = 0; i < _PB_N; i+=BLOCKSIZE)
                upc_memput(&globalPath[i], &localPath[i], BLOCKSIZE*sizeof(DATA_TYPE));
        }

        upc_barrier;
        if (ownerOfRow != MYTHREAD) {
            for(i = 0; i < _PB_N; i+=BLOCKSIZE)
                upc_memget(&localPath[i], &globalPath[i], BLOCKSIZE*sizeof(DATA_TYPE));
        }
        upc_barrier;
    }

    t_local = rtclock() - t_local;
    upc_barrier;

    double *t_localAll;
    t_localAll = (double *) malloc(THREADS*sizeof(double));
    t_global = bupc_allv_reduce(double, t_local, 0, UPC_MAX);
    t_localAll = bupc_allv_gather(double, t_local, 0, t_localAll);
    if (MYTHREAD==0) {
        char buffer[4096];
        int threadId;
        strcpy(buffer, "");
        for (threadId=0; threadId<THREADS; threadId++) {
            sprintf(buffer+strlen(buffer), "Node %d\n", threadId);
            sprintf(buffer+strlen(buffer), "Time: %lf\n", t_localAll[threadId]);
            sprintf(buffer+strlen(buffer), "-------\n");
        }
        sprintf(buffer+strlen(buffer), "SUMMARY\n");
        sprintf(buffer+strlen(buffer), "Maximum time spent across all nodes: %0.6lf s\n", t_global);
        sprintf(buffer+strlen(buffer), "-------");
        fprintf(stdout, "%s\n", buffer);
        fflush(stdout);
    }
    free(t_localAll);
    upc_barrier;

}


int main(int argc, char** argv)
{
    /* Retrieve problem size. */
    int n = N;

    /* Variable declaration/allocation. */
    POLYBENCH_2D_ARRAY_DECL(path, DATA_TYPE, N/THREADS, N, n/THREADS, n);

    shared [BLOCKSIZE] DATA_TYPE (*globalPath)[N];
    globalPath = upc_all_alloc(N/BLOCKSIZE, BLOCKSIZE*sizeof(DATA_TYPE));

    DATA_TYPE (*localPath)[N];
    localPath = malloc (N*sizeof(DATA_TYPE));

    /* Initialize array(s). */
    init_array (n, POLYBENCH_ARRAY(path));

    /* Start timer. */
    //polybench_start_instruments;

    /* Run kernel. */
    kernel_floyd_warshall (n, POLYBENCH_ARRAY(path), *globalPath, *localPath);

    /* Stop and print timer. */
    //polybench_stop_instruments;
    //polybench_print_instruments;

    /* Prevent dead-code elimination. All live-out data must be printed
       by the function call in argument. */
    polybench_prevent_dce(print_array(n, POLYBENCH_ARRAY(path)));

    /* Be clean. */
    POLYBENCH_FREE_ARRAY(path);
    if (MYTHREAD == 0)
        upc_free(globalPath);
    free(localPath);

    return 0;
}
